# This file is part of AdMincer,
# Copyright (C) 2019-present eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.


import os
import logging
import shutil
import xml.etree.ElementTree as etree
import glob

IMAGE_EXTENSIONS = ['.jpg', '.jpeg', '.png', '.gif']


def _get_names_file(path):
        """If there is a single .names file, return it."""
        names_files = glob.glob(os.path.join(path, '*.names'))
        if len(names_files) > 1:
            raise Exception('Multiple .names files in {}'
                            .format(path))
        if len(names_files) == 1:
            return names_files[0]
        return None


def _get_names(filename):
    """Open a filename and return a line-by-line list of the contents."""
    if os.path.exists(str(filename)):
        return [line.strip() for line in open(filename)]
    return []


def _append_unique(base, new):
    """Append any new, unique items to a base list."""
    for item in new:
        if item not in base:
            base.append(item)
    return base


def _write_names(names, names_dst):
    """Write the contents of a list to a file."""
    with open(names_dst, 'w+', newline='\n') as f:
        f.write('\n'.join(names))


def copy_move_images(image_dirs, output_dir, copy, move):
    for image_dir in image_dirs:
        for dirpath, dirnames, filenames in os.walk(image_dir):
            for fn in filenames:
                if any(fn.endswith(ext) for ext in IMAGE_EXTENSIONS):
                    if copy:
                        shutil.copy(os.path.join(dirpath, fn), output_dir)
                    elif move:
                        shutil.move(os.path.join(dirpath, fn), output_dir)


def convert_annotations(img_tag, names):
    """Find each ad in an img_tag and convert them to yolo-format annotations.

    Parameters
    ----------
    img_tag : xml.etree.ElementTree.Element
        CVAT-style tag for an image
    names : list
        An ordered list of labels, which describe boxes in images

    Returns
    -------
    (yolo_annotations, names) : tuple(str, list)
        yolo_annotations is a string containing all yolo-style annotations
            for the given img_tag, separated by '\n' newline characters.
        names is a list containing all passed-in names, plus any new labels.

    """
    width = int(img_tag.get('width'))
    height = int(img_tag.get('height'))

    yolo_annotations = []

    for box in img_tag.findall('box'):
        label = box.get('label')
        xmin = float(box.get('xtl'))
        ymin = float(box.get('ytl'))
        xmax = float(box.get('xbr'))
        ymax = float(box.get('ybr'))

        if label not in names:
            names.append(label)
        label_id = names.index(label)
        yolo_x = (xmin + ((xmax - xmin) / 2)) / width
        yolo_y = (ymin + ((ymax - ymin) / 2)) / height
        yolo_w = (xmax - xmin) / width
        yolo_h = (ymax - ymin) / height

        yolo_annotations.append(
            f'{label_id} {yolo_x:.6f} {yolo_y:.6f} {yolo_w:.6f} {yolo_h:.6f}'
        )

    return ('\n'.join(yolo_annotations), names)


def process_cvat_xml(xml_file, output_dir=None, move=False, copy=False):
    """Convert annotations from CVAT to YOLO.

    Parameters
    ----------
    xml_file : str
        Filename of the xml file containing the CVAT annotations
    output_dir : str
        Path where the YOLO annotation .txt files will be written
    move : bool
        If true, the images will be moved to the output_dir
    copy : bool
        If true, the images will be copied to the output_dir

    """
    if move and copy:
        raise ValueError('Choose either "copy" or "move", not both')

    cvat_xml = etree.parse(xml_file)

    # Is there a .names file in the output_dir?
    names_out = []
    names_file_out = None
    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)
        names_file_out = _get_names_file(output_dir)
        names_out = _get_names(names_file_out)

    # Is there a .names file alongside the XML file?
    names_file_up = _get_names_file(os.path.split(xml_file)[0])
    names_up = _get_names(names_file_up)

    # Is there a .names file alongside the images?
    image_folder = os.path.join(os.path.split(xml_file)[0],
                                os.path.split(cvat_xml.find('image')
                                              .get('name'))[0])
    names_file_in = _get_names_file(image_folder)
    names_in = _get_names(names_file_in)

    names = []
    names = _append_unique(names, names_out)
    names = _append_unique(names, names_up)
    names = _append_unique(names, names_in)

    image_dirs = set()

    for img_tag in cvat_xml.findall('image'):
        img_name = img_tag.get('name')
        image_path = os.path.join(os.path.split(xml_file)[0], img_name)
        image_dirs.add(os.path.dirname(image_path))

        if not os.path.exists(image_path):
            logging.warning('{} image cannot be found. Is `{}` image \
                            directory correct?'.
                            format(image_path, image_path))

        if output_dir is not None:
            yolo = os.path.basename(os.path.splitext(img_name)[0] + '.txt')
            yolo_dst = os.path.join(output_dir, yolo)

        else:
            yolo_dst = os.path.splitext(image_path)[0] + '.txt'

        yolo_annotations, names = convert_annotations(img_tag, names)

        with open(yolo_dst, 'w+', newline='\n') as yolo_file:
            yolo_file.write(yolo_annotations)

    # Write the class.names file to the output_dir or the image paths, and
    # overwrite any existing .names files
    if output_dir is not None:
        if names_file_out:
            _write_names(names, names_file_out)
        else:
            _write_names(names, os.path.join(output_dir, 'class.names'))
    else:
        for image_dir in image_dirs:
            _write_names(names, os.path.join(image_dir, 'class.names'))

    if names_file_up:
        _write_names(names, names_file_up)

    if names_file_in:
        _write_names(names, names_file_in)

    if output_dir is not None and copy or move:
        copy_move_images(image_dirs, output_dir, copy, move)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Common fixtures."""

from PIL import Image
import pytest
import py


class Colors:
    """Colors for composing images and checking output images."""
    RED = (255, 0, 0)
    GREEN = (0, 255, 0)
    ORANGE = (255, 165, 0)
    BLUE = (0, 0, 255)
    PINK = (255, 128, 128)
    WHITE = (255, 255, 255)
    BLACK = (0, 0, 0)


def make_image(path, width, height, color=Colors.WHITE):
    """Make a dummy image for testing."""
    Image.new('RGB', (width, height), color).save(str(path))


@pytest.fixture()
def logs_dir():
    """Directory with expected logs."""
    return py.path.local(__file__).dirpath().join('data', 'logs')


@pytest.fixture()
def fragdir(tmpdir):
    """Directory with some image fragments."""
    ret = tmpdir.mkdir('fragdir')
    make_image(ret.join('red.gif'), 10, 20, Colors.RED)
    make_image(ret.join('green.png'), 30, 40, Colors.GREEN)
    make_image(ret.mkdir('subdir').join('orange.jpg'), 50, 60, Colors.ORANGE)
    ret.join('other.sh').write('# random unrelated file')
    return ret


@pytest.fixture()
def regdir(tmpdir):
    """Directory with some images with marked regions."""
    ret = tmpdir.mkdir('regdir')
    # White square with 2 regions (see below for regions).
    make_image(ret.join('white2.png'), 100, 100, Colors.WHITE)
    # Blue rectangle with 3 regions.
    make_image(ret.join('blue3.png'), 200, 100, Colors.BLUE)
    # Black square with no regions.
    make_image(ret.join('black.gif'), 100, 100, Colors.BLACK)
    ret.join('index.csv').write('\n'.join([
        # Some rows have text in quotes while others don't. We should be
        # compatible with both styles.
        '"image","xmin","ymin","xmax","ymax","label"',
        '"white2.png",10.34,20.12,19.79,39.9,"r1"',
        'white2.png,50,50,80,90,r2',
        'blue3.png,10,10,20,20,r1',
        'blue3.png,40,10,50,20,r1',
        'blue3.png,10,30,60,90,r3',
    ]))
    return ret


@pytest.fixture()
def regdir_var(tmpdir, regdir):
    """Directory with images and regions in different formats."""
    ret = tmpdir.join('regdir_var')
    regdir.copy(ret)
    ret.join('index.csv').write('\n'.join([
        'image,xmin,ymin,xmax,ymax,label',
        'black.gif,10,10,20,20,r1',
    ]))
    ret.join('region.names').write('\n'.join(['r1', 'r2', 'r3', '']))
    ret.join('white2.txt').write('\n'.join([
        '0 0.05 0.2 0.2 0.5',  # Should be clipped from top and left.
        '1 0.85 0.9 0.4 0.4',  # Should be clipped from bottom and right.
        '',
    ]))
    ret.join('blue3.txt').write('\n'.join([
        '0 0.075 0.15 0.05 0.1',
        '1 0.225 0.15 0.05 0.1',
        '2 0.175 0.6 0.25 0.6',
    ]))
    return ret


@pytest.fixture()
def regdir_with_clipping(tmpdir):
    """Directory with some images with marked regions."""
    ret = tmpdir.mkdir('regdir_with_clipping')
    # Red square with 3 regions, all in need of clipping
    make_image(ret.join('red3.png'), 100, 90, Colors.BLACK)
    ret.join('index.csv').write('\n'.join([
        # Some rows have text in quotes while others don't. We should be
        # compatible with both styles.
        '"image","xmin","ymin","xmax","ymax","label"',
        'red3.png,-10,80,20,120,r1',
        'red3.png,90,-20,130,20,r1',
        'red3.png,0,20,0,50,r1',
        'red3.png,20,0,30,0,r1',
    ]))
    return ret


@pytest.fixture()
def regdir_pandas(tmpdir):
    """Directory with images and regions in pandas format."""
    ret = tmpdir.mkdir('regdir_pandas')
    make_image(ret.join('black.gif'), 100, 100, Colors.BLACK)
    make_image(ret.join('blue3.png'), 200, 100, Colors.BLUE)
    make_image(ret.join('white2.png'), 100, 100, Colors.WHITE)
    ret.join('regions.csv').write('\n'.join([
        ',image,xmin,ymin,xmax,ymax,label',
        '0,black.gif,10,10,20,20,r1',
        '1,blue3.png,10,10,20,20,r2',
        '2,white2.png,50,50,80,90,r3',
    ]))
    return ret


def _char_to_color(c):
    """Convert character to an RGB color."""
    code = ord(c)
    r = (code & 0b11) * 64
    g = ((code >> 2) & 0b11) * 64
    b = ((code >> 4) & 0b111) * 32
    return (r, g, b)


def _color_to_char(color):
    """Convert RGB color to a character."""
    r, g, b = color
    low = int(round(r / 64))
    mid = int(round(g / 64)) * 4
    hi = int(round(b / 32)) * 16
    code = low + mid + hi
    return chr(code)


def chars_to_image(chars, cell_size=8, width=4, height=4):
    """Make an image based on a string of characters.

    Parameters
    ----------
    chars : str
        Characters that define the colors of cells.
    cell_size : int
        Height and width of each cell (which will be colored the same).
    width : int
        Width of the image (in cells).
    height : int
        Height of the image (in cells).

    Returns
    -------
    img : PIL.Image
        Resulting image.

    """
    colors = [_char_to_color(c) for c in chars]
    img = Image.new('RGB', (cell_size * width, cell_size * height))
    img.putdata([
        colors[row * width + col]
        for row in range(height)
        for y in range(cell_size)
        for col in range(width)
        for x in range(cell_size)
    ])
    return img


def image_to_chars(image, cell_size=8, width=4, height=4):
    """Convert image produced by chars_to_image back to sting of characters.

    Parameters
    ----------
    image : PIL.Image
        Image to convert.
    cell_size : int
        Height and width of each cell (which will be colored the same).
    width : int
        Width of the image (in cells).
    height : int
        Height of the image (in cells).

    Returns
    -------
    chars : str
        The characters for the image.

    """
    assert image.size == (cell_size * width, cell_size * height)
    offset = cell_size // 2
    return ''.join(
        _color_to_char(image.getpixel((col * cell_size + offset,
                                       row * cell_size + offset)))
        for row in range(height)
        for col in range(width)
    )


# Characters to make images from.
STRINGS = [
    'just some random',
    'chars here to ma',
    'e fragments from',
    'f23sj09xczvnm,we',
]


def _make_needles():
    """Make 4 distinct 32x32 images to use as pieces of big images."""
    return [chars_to_image(chars) for chars in STRINGS]


def _make_haystack(needles):
    """Make 4 images that contain some of the needles."""
    needle_placements = [
        [0, 1, 2],
        [2, 3],
        [0, 2],
        [1, 3],
    ]
    filler = chars_to_image('for the bckground')

    ret = []
    for np in needle_placements:
        img = Image.new('RGB', (100, 200))
        for x in [0, 30, 60, 80]:
            for y in [0, 30, 60, 90, 120, 150, 170]:
                img.paste(filler, (x, y, x + 32, y + 32))
        for i, n in enumerate(np):
            needle = needles[n]
            img.paste(needle, (i * 20, i * 50, i * 20 + 32, i * 50 + 32))
        ret.append(img)

    return ret


@pytest.fixture(scope='session')
def fragfind_dir(tmpdir_factory):
    """Directory that contains files for testing the fragment search."""
    ret = tmpdir_factory.mktemp('fragdir')
    needles_dir = ret.mkdir('needles')
    haystack_dir = ret.mkdir('haystack')

    needles = _make_needles()

    for i, needle in enumerate(needles):
        path = needles_dir.join('{}.png'.format(i))
        print('saving to {}'.format(path))
        needle.save(str(path))
    
    haystack = _make_haystack(needles)

    for i, img in enumerate(haystack):
        # Save these ones as JPG to test inexact matching.
        path = haystack_dir.join('{}.jpg'.format(i))
        img.save(str(path))

    haystack_dir.join('regions.csv').write('\n'.join([
        'image,xmin,ymin,xmax,ymax,label',
        '0.jpg,0,0,10,10,r1',
        '1.jpg,0,0,10,10,r2',
        '2.jpg,0,0,17,10,r1',
        '3.jpg,0,0,10,10,r1',
    ]))

    return ret


SLICE_STRINGS = [
    '123456789',
    'abcdefghi',
    'jklmnopqrstu',
    'ABCDEFGHIJKLMNOPQRSTUVWX',
]


@pytest.fixture(scope='session')
def slice_source(tmpdir_factory):
    """Source directory for tests of slicing functionality."""
    ret = tmpdir_factory.mktemp('slice_source')

    for i in range(len(SLICE_STRINGS)):
        chars = SLICE_STRINGS[i]
        image = chars_to_image(chars, 10, 3, len(chars) // 3)
        image.save(str(ret.join('{}.png'.format(i))))

    ret.join('regions.csv').write('\n'.join([
        'image,xmin,ymin,xmax,ymax,label',
        '0.png,0,0,10,10,r1',
        '0.png,20,10,30,30,r2',
        '2.png,0,0,10,10,r1',
        '2.png,20,0,30,20,r3',
        '2.png,20,30,30,40,r1',
        '3.png,20,10,30,30,r1',
        '3.png,20,60,30,80,r2',
    ]))

    return ret


@pytest.fixture()
def cvat_dir():
    """Directory with CVAT-style annotation files."""
    return py.path.local(__file__).dirpath().join('data', 'cvat_xmls')


@pytest.fixture()
def convert_src(tmpdir_factory, cvat_dir):
    """Source directory for tests of converting functionality."""
    ret = tmpdir_factory.mktemp('convert_source')
    task11_images = ret.mkdir('task11')
    task12_images = ret.mkdir('task12')

    cvat_dir.join('14_task11.xml').copy(ret.join('14_task11.xml'))
    cvat_dir.join('15_task12.xml').copy(ret.join('15_task12.xml'))

    make_image(task11_images.join('1_cut_0.png'), 1351, 1351, Colors.WHITE)
    make_image(task11_images.join('1_cut_1.png'), 1351, 1351, Colors.BLUE)
    make_image(task11_images.join('no_ads.png'), 1351, 1351, Colors.RED)
    make_image(task12_images.join('5_cut_1.png'), 1351, 1351, Colors.BLACK)
    make_image(task12_images.join('2_cut_0.png'), 1351, 1351, Colors.RED)

    return ret


@pytest.fixture()
def convert_src_names(tmpdir_factory, cvat_dir):
    """Directory with CVAT annotations and .names files."""
    ret = tmpdir_factory.mktemp('convert_source')
    task11_images = ret.mkdir('task11')
    task12_images = ret.mkdir('task12')

    task11_images.join('class.names').write('bannerad\ntextad\n')
    task12_images.join('class.names').write('textad\nbannerad\npopupad')

    cvat_dir.join('14_task11.xml').copy(ret.join('14_task11.xml'))
    cvat_dir.join('15_task12.xml').copy(ret.join('15_task12.xml'))

    make_image(task11_images.join('1_cut_0.png'), 1351, 1351, Colors.WHITE)
    make_image(task11_images.join('1_cut_1.png'), 1351, 1351, Colors.BLUE)
    make_image(task12_images.join('5_cut_1.png'), 1351, 1351, Colors.BLACK)
    make_image(task12_images.join('2_cut_0.png'), 1351, 1351, Colors.RED)

    return ret


@pytest.fixture()
def bad_convert_src(tmpdir_factory, cvat_dir):
    """Source directory for tests of converting functionality."""
    ret = tmpdir_factory.mktemp('convert_source')
    task11_images = ret.mkdir('task11')

    cvat_dir.join('14_task11.xml').copy(ret.join('14_task11.xml'))
    make_image(task11_images.join('1_cut_0.png'), 1351, 1351, Colors.WHITE)

    return ret



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Sanity checks for the test suite."""

from conftest import chars_to_image, image_to_chars, STRINGS


def test_image_chars():
    for chars in STRINGS:
        assert image_to_chars(chars_to_image(chars)) == chars



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Searching for images based on regions and fragments."""

import functools
import os

import cv2


class Query:
    """Base class for queries."""

    difficulty = 1  # How much compute does this query need?

    def is_match(self, image_path, data):
        """Does this image match the query?"""
        raise NotImplementedError()  # pragma: no cover 

    def __str__(self):
        return self.format_string.format(self)


class RegionQuery(Query):
    """Query for regions of specific type and size."""

    format_string = (
        'RegionQuery({0.region_type}, '
        '{0.min_width} <= width <= {0.max_width}, '
        '{0.min_height} <= height <= {0.max_height})'
    )

    def __init__(self, region_type, width, height, tolerance=25):
        self.region_type = region_type
        multiplier = 1 + tolerance / 100
        self.max_width = width * multiplier
        self.min_width = width / multiplier
        self.max_height = height * multiplier
        self.min_height = height / multiplier

    def is_match(self, image_path, data):
        """Is there a matching (within tolerance) region?"""
        if not isinstance(data, list):
            raise Exception('Region queries require a region index')
        for x1, y1, x2, y2, region_type in data:
            if (
                region_type == self.region_type
                and self.min_width <= x2 - x1 <= self.max_width
                and self.min_height <= y2 - y1 <= self.max_height
            ):
                return True
        return False


class FragmentQuery(Query):
    """Query for a fragment image."""

    difficulty = 10
    format_string = 'FragmentQuery({0.frag_path}, tolerance={0.tolerance})'

    def __init__(self, frag_path, tolerance=10):
        self.frag_path = frag_path
        self.tolerance = tolerance / 100

    @staticmethod
    @functools.lru_cache(maxsize=1)
    def _load_image(image_path):
        """Cache images between queries."""
        with open(image_path, 'r'):
            pass  # Check that we can read the image (OpenCV doesn't).
        return cv2.imread(image_path)

    def is_match(self, image_path, data):
        """Is the best template match below tolerance?"""
        if not hasattr(self, 'frag'):
            self.frag = FragmentQuery._load_image(self.frag_path)
        image = FragmentQuery._load_image(image_path)
        res = cv2.matchTemplate(self.frag, image, cv2.TM_SQDIFF_NORMED)
        best = res.min()
        return best < self.tolerance


def find(index, queries):
    """Find images in the index that match all the queries."""
    queries = sorted(queries, key=lambda q: q.difficulty)
    for name in sorted(index):
        matched_queries = set()
        path = os.path.join(index.root_path, name)
        data = index[name]
        if all(query.is_match(path, data) for query in queries):
            yield name



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Common utilities."""

import argparse

import admincer.find as fnd


def region(reg):
    """Parser for region queries."""
    params = reg.split(':')
    if len(params) not in {2, 3}:
        raise argparse.ArgumentTypeError('expected format: TYPE:WxH[:T]')
    width, height = params[1].split('x')
    params[1:2] = float(width), float(height)
    params[-1] = float(params[-1])
    return fnd.RegionQuery(*params)


def fragment(frg):
    """Parser for fragment queries."""
    params = frg.split(':')
    if len(params) == 2:
        params[1] = float(params[1])
    elif len(params) != 1:
        raise argparse.ArgumentTypeError('expected format: PATH[:T]')
    return fnd.FragmentQuery(*params)


def format_box(box):
    """Format box description for display."""
    x0, y0, x1, y1 = box
    return '({}, {}: {}x{})'.format(x0, y0, x1 - x0, y1 - y0)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Indexing images in directories."""

import logging
import os

from PIL import Image


def extract_regions(reg_index, region_type):
    """Return the contents of the regions as images.

    Parameters
    ----------
    reg_index : Index
        Index of source images and their regions.
    region_type : str
        Region type to extract.

    Returns
    -------
    regions : iterable(Image)
        Extracted regions as PIL images.

    """
    for rel_path, regions in sorted(reg_index.items()):
        boxes = [r[:4] for r in regions if r[4] == region_type]
        if not boxes:
            continue
        logging.info('### %s', rel_path)
        image_path = os.path.join(reg_index.root_path, rel_path)
        image = Image.open(image_path)
        for box in boxes:
            yield box, image.crop(box)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for region extraction."""

import pytest

import admincer.index as idx
import admincer.extract as ex

from conftest import Colors as cc


@pytest.mark.parametrize('region_type, expected_params', [
    ('r0', []),
    ('r1', [(10, 20, cc.WHITE), (10, 10, cc.BLUE), (10, 10, cc.BLUE)]),
    ('r2', [(30, 40, cc.WHITE)]),
])
def test_extract(regdir, region_type, expected_params):
    ri = idx.reg_index(str(regdir))
    extracted_images = [
        image for box, image in ex.extract_regions(ri, region_type)
    ]
    assert len(extracted_images) == len(expected_params)
    extracted_params = {
        image.size + (image.getpixel((0, 0)),)
        for image in extracted_images
    }
    assert extracted_params == set(expected_params)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for image indexing."""

import pytest

import admincer.index as idx


@pytest.mark.parametrize('constructor', [idx.frag_index, idx.some_index])
def test_frag_index(fragdir, constructor):
    """Create an index of a directory with some fragments."""
    path = str(fragdir)
    fi = constructor(path)
    assert fi.root_path == path
    assert fi == {
        'red.gif': (10, 20),
        'green.png': (30, 40),
        'subdir/orange.jpg': (50, 60),
    }


def test_frag_add(fragdir):
    """Add image to fragment index."""
    fi = idx.frag_index(str(fragdir))
    fi.add_image('foo.png', fi.load_image('red.gif'))


@pytest.mark.parametrize('constructor', [idx.reg_index, idx.some_index])
def test_reg_index(regdir, constructor):
    """Create an index of a directory with images that have marked regions."""
    path = str(regdir)
    ri = constructor(path)
    assert ri.root_path == path
    assert ri == {
        'white2.png': [
            (10, 20, 20, 40, 'r1'),
            (50, 50, 80, 90, 'r2'),
        ],
        'blue3.png': [
            (10, 10, 20, 20, 'r1'),
            (40, 10, 50, 20, 'r1'),
            (10, 30, 60, 90, 'r3'),
        ],
        'black.gif': [],
    }


def test_reg_index_clipping(regdir_with_clipping):
    """Create an index of a directory with images that have marked regions."""
    path = str(regdir_with_clipping)
    ri = idx.reg_index(path)
    assert ri.root_path == path
    assert ri == {
        'red3.png': [
            (0, 80, 20, 90, 'r1'),
            (90, 0, 100, 20, 'r1'),
        ],

    }


def test_reg_index_yolo(regdir_var):
    """Region loading from YOLO-format files."""
    path = str(regdir_var)
    ri = idx.reg_index(path)
    assert ri.region_types == ['r1', 'r2', 'r3']
    assert ri == {
        'white2.png': [
            (0, 0, 15, 45, 'r1'),
            (65, 70, 100, 100, 'r2'),
        ],
        'blue3.png': [
            (10, 10, 20, 20, 'r1'),
            (40, 10, 50, 20, 'r2'),
            (10, 30, 60, 90, 'r3'),
        ],
        'black.gif': [
            (10, 10, 20, 20, 'r1'),
        ],
    }


def test_reg_index_yolo_nonames(regdir_var):
    """Region loading from YOLO-format files without .names mapping."""
    regdir_var.join('region.names').remove()
    regdir_var.join('index.csv').remove()
    path = str(regdir_var)
    ri = idx.reg_index(path)
    assert ri.region_types == ['0', '1', '2']  # Autogenerated.
    assert ri == {
        'white2.png': [
            (0, 0, 15, 45, '0'),
            (65, 70, 100, 100, '1'),
        ],
        'blue3.png': [
            (10, 10, 20, 20, '0'),
            (40, 10, 50, 20, '1'),
            (10, 30, 60, 90, '2'),
        ],
        'black.gif': [],
    }


def test_reg_index_pandas(regdir_pandas):
    """sth."""
    path = str(regdir_pandas)
    ri = idx.reg_index(path)
    assert ri.region_types == ['r1', 'r2', 'r3']
    assert ri == {
        'black.gif': [(10, 10, 20, 20, 'r1')],
        'blue3.png': [(10, 10, 20, 20, 'r2')],
        'white2.png': [(50, 50, 80, 90, 'r3')],
    }


def test_reg_index_2names(regdir_var):
    """2 .names files -> error."""
    regdir_var.join('other.names').write('foo\n')
    path = str(regdir_var)
    with pytest.raises(Exception) as info:
        idx.reg_index(path)
    assert str(info.value).startswith('Multiple .names files')


def test_reg_index_csvtxt(regdir_var):
    """Regions for the same image in both .csv and .txt files -> error."""
    regdir_var.join('index.csv').write('\n'.join([
        'image,xmin,ymin,xmax,ymax,label',
        'white2.png,10,10,10,10,r1',
    ]))
    path = str(regdir_var)
    with pytest.raises(Exception) as info:
        idx.reg_index(path)
    assert str(info.value).startswith('Regions specified twice')


def test_reg_index_nonames(regdir_var):
    """No .names file and regions in both .csv and .txt -> error."""
    regdir_var.join('region.names').remove()
    path = str(regdir_var)
    with pytest.raises(Exception) as info:
        idx.reg_index(path)
    assert str(info.value).startswith('A .names file is required')



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Place fragments into regions of an image."""

import collections
import csv
import logging
import os
import random

from PIL import Image

import admincer.util as util


# How much we can scale fragments to make them fit.
MIN_SCALE = 0.8
MAX_SCALE = 1.25


def get_pad_color(image, region):
    """Determine the color to use for padding fragments placed in this region.

    The return color is the most common color in the bottom and right
    boundaries of the region.

    Parameters
    ----------
    image : Image
        Source image.
    region : (int, int, int, int)
        Region coordinates (x1, y1, x2, y2).

    Returns
    -------
    color : (int, int, int)
        Background color as an RGB triple.

    """
    points = (
        [(x, region[3] - 1) for x in range(region[0], region[2])] +
        [(region[2] - 1, y) for y in range(region[1], region[3])]
    )
    values = [image.getpixel(p) for p in points]
    counts = collections.defaultdict(int)
    for v in values:
        counts[v] += 1
    return max((c, v) for v, c in counts.items())[1]


def _place(image, fragment, region):
    """Place fragment into the region of the image that has the same size."""
    if fragment.mode == 'RGBA':
        image.paste(fragment, region, fragment)
    else:
        image.paste(fragment, region)


def place_fragment(image, fragment, region, resize_mode='scale'):
    """Place fragment into the image.

    Parameters
    ----------
    image : Image
        Image onto which the fragment will be placed.
    fragment : Image
        The fragment to place.
    region : (int, int, int, int)
        Region coordinates (x1, y1, x2, y2).
    resize_mode : str
        A flag that controls the way the image size is adjusted to the size of
        the region. One of 'scale', 'pad' or 'crop'.

    """
    x1, y1, x2, y2 = region
    size = x2 - x1, y2 - y1

    if fragment.size != size:
        if resize_mode == 'scale':
            fragment = fragment.resize(size, Image.HAMMING)
        elif resize_mode == 'pad':
            pad_color = get_pad_color(image, region)
            padded_fragment = Image.new('RGB', size, pad_color)
            size = tuple(min(size[i], fragment.size[i]) for i in range(2))
            fragment = fragment.crop((0, 0) + size)
            _place(padded_fragment, fragment, (0, 0) + size)
            fragment = padded_fragment
        elif resize_mode == 'crop':
            size = tuple(min(size[i], fragment.size[i]) for i in range(2))
            region = (x1, y1, x1 + size[0], y1 + size[1])
            fragment = fragment.crop((0, 0) + size)
        else:
            raise ValueError('Resize mode must be "scale", "pad" or "crop"')

    _place(image, fragment, region)


class Recipe:
    """Fragments placement recipe."""

    def __init__(self, image_path, placements):
        """Constructor.

        Parameters
        ----------
        image_path : str
            Path to the main image.
        placements : [(fragment_path, x1, y1, x2, y2)]
            List of fragment placements.

        """
        self.image_path = image_path
        self.placements = placements

    def render(self, target_path):
        """Render the result of this recipe.

        Parameters
        ----------
        target_path : str
            Path where the result of the rendering will be written.

        """
        logging.info('- source image: %s', self.image_path)
        image = Image.open(self.image_path)
        for fragment_path, region, resize_mode in self.placements:
            f_region = util.format_box(region)
            if fragment_path is None:
                logging.info('- no fragment for region: %s', f_region)
                continue
            logging.info('- fragment from: %s into region: %s',
                         fragment_path, f_region)
            fragment = Image.open(fragment_path)
            place_fragment(image, fragment, region, resize_mode)
        logging.info('- save as: %s', target_path)
        image.save(target_path)


class FragmentPicker:
    """Picks fragments from one or more fragment indices"""

    def __init__(self, fragindex_map):
        self.fragindex_map = fragindex_map

    def pick(self, width, height, region_type):
        """Pick a fragment that matches the region type and size.

        Parameters
        ----------
        width : int
            Width of the region.
        height : int
            Height of the region.
        region_type : str
            Type of the region.

        Returns
        -------
        fragment_path : str or None
            Full path to the fragment image or None if there are no matching
            fragments.

        """
        indices = set()
        indices.update(
            self.fragindex_map.get(region_type, []),
            self.fragindex_map.get(None, []),
        )
        candidates = [
            os.path.join(index.root_path, image)
            for index in indices
            for image, (w, h) in index.items()
            if (w * MIN_SCALE <= width <= w * MAX_SCALE and
                h * MIN_SCALE <= height <= h * MAX_SCALE)
        ]
        if candidates:
            return random.choice(candidates)


def gen_recipes(regindex, fragindices, resize_modes=None):
    """Generate fragment placement recipes.

    Parameters
    ----------
    regindex : Index
        Index of source images and regions in them.
    fragindices : [(str, [Index])]
        Sequence of pairs: region label, list of fragment indices.

    Returns
    -------
    recipes : generator of Recipe
        Fragments placement recipes.

    Note: the placements follow the order of region labels in `fragindices`.

    """
    if resize_modes is None:
        resize_modes = {}
    if None not in resize_modes:
        resize_modes[None] = 'scale'

    fragindex_map = collections.OrderedDict()
    for label, indices in fragindices:
        fragindex_map.setdefault(label, []).extend(indices)
    fp = FragmentPicker(fragindex_map)

    images = list(regindex)
    while True:
        image = random.choice(images)
        regions = regindex[image]
        image_path = os.path.join(regindex.root_path, image)
        placements = [
            (
                fp.pick(x2 - x1, y2 - y1, region_type),
                (x1, y1, x2, y2),
                resize_modes.get(region_type, resize_modes[None]),
            )
            for rt in fragindex_map
            for (x1, y1, x2, y2, region_type) in regions
            if region_type == rt
        ]
        yield Recipe(image_path, placements)


def _write_regions_csv(csv_path, regions_dict):
    """Write regions from a dictionary into a CSV file."""
    with open(csv_path, 'wt', encoding='utf-8') as f:
        csv_w = csv.writer(f)
        csv_w.writerow(('image', 'xmin', 'ymin', 'xmax', 'ymax', 'label'))

        for image_name, regions in regions_dict.items():
            for region in regions:
                csv_w.writerow((image_name,) + region)


def render_recipes(regindex, recipes, target_dir):
    """Render multiple recipes to target_dir.

    Parameters
    ----------
    regindex : Index
        Index of source images and regions in them.
    recipes : iterable of Recipe
        The recipes to render.

    """
    regions = {}
    os.makedirs(target_dir, exist_ok=True)

    for i, recipe in enumerate(recipes):
        out_name = '{:05d}.png'.format(i)
        out_path = os.path.join(target_dir, out_name)
        logging.info('## %s', out_name)
        recipe.render(out_path)
        regions[out_name] = regindex[os.path.basename(recipe.image_path)]

    _write_regions_csv(os.path.join(target_dir, 'regions.csv'), regions)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for image search."""

from PIL import Image
import pytest

import admincer.find as fnd
import admincer.index as idx


@pytest.fixture(scope='session')
def index():
    return idx.RegionIndex('', {
        'foo': [
            (10, 10, 20, 30, '1'),
            (0, 0, 50, 60, '2'),
            (5, 5, 10, 10, '3'),
        ],
        'bar': [
            (10, 10, 20, 30, '1'),
            (0, 0, 50, 70, '2'),
        ],
        'baz': [
            (0, 0, 50, 60, '2'),
            (5, 5, 10, 10, '3'),
        ],
    })


@pytest.mark.parametrize('params,result', [
    (('foo', 100, 200),
     'RegionQuery(foo, 80.0 <= width <= 125.0, 160.0 <= height <= 250.0)'),
    (('bar', 100, 200, 100),
     'RegionQuery(bar, 50.0 <= width <= 200.0, 100.0 <= height <= 400.0)'),
])
def test_region_query(params, result):
    """Test the RegionQuery constructor."""
    rq = fnd.RegionQuery(*params)
    assert str(rq) == result


@pytest.mark.parametrize('queries,expected', [
    ([fnd.RegionQuery('1', 10, 20)], ['bar', 'foo']),
    ([fnd.RegionQuery('1', 10, 20),
      fnd.RegionQuery('2', 50, 50)], ['foo']),
    ([fnd.RegionQuery('1', 10, 20),
      fnd.RegionQuery('2', 50, 50)], ['foo']),
])
def test_find_region(index, queries, expected):
    found = list(fnd.find(index, queries))
    assert found == expected


@pytest.mark.parametrize('needles,expected', [
    (['0.png'], ['0.jpg', '2.jpg']),
    (['1.png'], ['0.jpg', '3.jpg']),
    (['2.png'], ['0.jpg', '1.jpg', '2.jpg']),
    (['2.png', '3.png'], ['1.jpg']),
])
def test_find_fragment(fragfind_dir, needles, expected):
    """Test finding images by fragments."""
    index = idx.some_index(str(fragfind_dir.join('haystack')))
    queries = [
        fnd.FragmentQuery(str(fragfind_dir.join('needles').join(needle)))
        for needle in needles
    ]
    found = list(fnd.find(index, queries))
    assert found == expected


def test_find_fragment_and_region(fragfind_dir):
    """Test finding images by fragments and regions."""
    index = idx.some_index(str(fragfind_dir.join('haystack')))
    queries = [
        fnd.RegionQuery('r1', 10, 10),
        fnd.FragmentQuery(str(fragfind_dir.join('needles').join('2.png')))
    ]
    found = list(fnd.find(index, queries))
    assert found == ['0.jpg']



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for placing fragments into regions."""

import os

from PIL import Image, ImageDraw
import pytest

import admincer.index as idx
import admincer.place as pl

from conftest import Colors as cc

# Fragment indices.
A = idx.FragmentIndex('a', {
    '1': (10, 10),
    '2': (20, 20),
})
B = idx.FragmentIndex('b', A)
C = idx.FragmentIndex('c', {
    '1/1': (10, 10),
    '1/2': (10, 10),
    '2/1': (20, 20),
    '2/2': (20, 20),
})
D = idx.FragmentIndex('d', C)
E = idx.FragmentIndex('e', {
    '1': (9, 9),
    '2': (11, 11),
    '3': (18, 22),
    '4': (22, 18),
})

# Images with regions.
X = idx.RegionIndex('x', {
    '1': [
        (10, 10, 20, 20, 'a'),
        (20, 20, 30, 30, 'b'),
        (30, 30, 50, 50, 'a'),
        (50, 50, 70, 70, 'b'),
    ],
})
Y = idx.RegionIndex('y', {
    '1': [(10, 10, 20, 20, 'a')],
    '2': [(10, 10, 20, 20, 'b')],
    '3': [(10, 10, 30, 30, 'a')],
    '4': [(10, 10, 30, 30, 'b')],
})


def test_fragment_picker():
    fp = pl.FragmentPicker({'a': [A], 'b': [B]})
    assert fp.pick(10, 10, 'a') == 'a/1'
    assert fp.pick(10, 10, 'b') == 'b/1'
    assert fp.pick(20, 20, 'a') == 'a/2'
    assert fp.pick(20, 20, 'b') == 'b/2'
    assert fp.pick(20, 20, 'c') is None


def test_fragment_picker_randomness():
    fp = pl.FragmentPicker({None: [A, B]})
    picks = [fp.pick(10, 10, 'a') for _ in range(100)]
    # Check that each eligible was picked between 20 and 80 times out of 100.
    for i in 'ab':
        assert 20 < len([p for p in picks if p == i + '/1']) < 80


def test_gen_recipes():
    rg = pl.gen_recipes(X, [('a', [A]), ('b', [B])], {'b': 'pad'})
    recipe = next(rg)
    assert recipe.image_path == 'x/1'
    assert recipe.placements == [
        ('a/1', (10, 10, 20, 20), 'scale'),
        ('a/2', (30, 30, 50, 50), 'scale'),
        ('b/1', (20, 20, 30, 30), 'pad'),
        ('b/2', (50, 50, 70, 70), 'pad'),
    ]
    # Reverse order of fragment indices.
    rg = pl.gen_recipes(X, [('b', [B]), ('a', [A])])
    recipe = next(rg)
    assert [p[0] for p in recipe.placements] == ['b/1', 'b/2', 'a/1', 'a/2']


def test_gen_recipes_randomness():
    rg = pl.gen_recipes(Y, [('a', [C]), ('b', [D])])
    recipes = [next(rg) for _ in range(100)]
    variety = {
        '{}-{}'.format(recipe.image_path, recipe.placements[0][0])
        for recipe in recipes
    }
    assert variety == {
        'y/1-c/1/1', 'y/1-c/1/2',
        'y/2-d/1/1', 'y/2-d/1/2',
        'y/3-c/2/1', 'y/3-c/2/2',
        'y/4-d/2/1', 'y/4-d/2/2',
    }


def test_close_enough():
    """Try placing fragments into imprecisely matching regions."""
    rg = pl.gen_recipes(X, [('a', [E]), ('b', [E])])
    recipe = next(rg)
    assert len(recipe.placements) == 4


@pytest.fixture()
def wb_100x100():
    """White 100x100 with blue lines on left and top."""
    image = Image.new('RGB', (100, 100), cc.WHITE)
    draw = ImageDraw.Draw(image)
    draw.rectangle([(0, 0), (2, 100)], cc.BLUE)
    draw.rectangle([(0, 0), (100, 2)], cc.BLUE)
    return image


@pytest.mark.parametrize(
    'region,resize_mode,in_corner,out_corner,plc_width,plc_height', [
    ((0, 0, 10, 10), 'scale', (5, 5), (10, 10), 10, 10),
    ((0, 0, 14, 14), 'scale', (7, 7), (14, 14), 14, 14),
    ((0, 0, 8, 8), 'scale', (4, 4), (8, 8), 8, 8),
    ((0, 0, 10, 10), 'crop', (5, 5), (10, 10), 10, 10),  # Same as scale.
    ((0, 0, 14, 14), 'crop', (5, 5), (10, 10), 10, 10),
    ((0, 0, 8, 8), 'crop', (5, 5), (8, 8), 8, 8),
    ((0, 0, 10, 10), 'pad', (5, 5), (10, 10), 10, 10),  # Same as scale.
    ((0, 0, 14, 14), 'pad', (5, 5), (10, 10), 14, 14),
    ((0, 0, 8, 8), 'pad', (5, 5), (8, 8), 8, 8),
])
def test_fragment_placement(
    wb_100x100,
    region, resize_mode, in_corner, out_corner, plc_width, plc_height,
):
    """Test placement of fragments."""
    image = wb_100x100
    fragment = Image.new('RGB', (10, 10), cc.GREEN)
    draw = ImageDraw.Draw(fragment)
    draw.rectangle([(0, 0), (5, 5)], cc.RED)

    pl.place_fragment(image, fragment, region, resize_mode)
    rg_x, rg_y = in_corner
    gw_x, gw_y = out_corner
    # Check that specific points of placed fragment are where they should be.
    assert image.getpixel((rg_x - 2, rg_y - 2)) == cc.RED
    assert image.getpixel((rg_x + 2, rg_y - 2)) == cc.GREEN
    assert image.getpixel((rg_x - 2, rg_y + 2)) == cc.GREEN
    assert image.getpixel((gw_x - 2, gw_y - 2)) == cc.GREEN
    assert image.getpixel((gw_x + 2, gw_y - 2)) == cc.WHITE
    assert image.getpixel((gw_x - 2, gw_y + 2)) == cc.WHITE
    # Check that overall modified area has the right size.
    assert image.getpixel((plc_width - 1, 1)) != cc.BLUE
    assert image.getpixel((plc_width + 1, 1)) == cc.BLUE
    assert image.getpixel((1, plc_height - 1)) != cc.BLUE
    assert image.getpixel((1, plc_height + 1)) == cc.BLUE


@pytest.mark.parametrize('resize_mode,expect_colors', [
    ('scale', [cc.RED, cc.RED, cc.PINK,
               cc.RED, cc.RED, cc.BLUE, cc.WHITE]),
    ('crop',  [cc.RED, cc.PINK, cc.WHITE,
               cc.BLUE, cc.WHITE, cc.BLUE, cc.WHITE]),
    ('pad',   [cc.RED, cc.PINK, cc.WHITE,
               cc.WHITE, cc.WHITE, cc.WHITE, cc.WHITE]),
])
def test_transparent_placement(wb_100x100, resize_mode, expect_colors):
    """Place a partially transparent fragment in different resize modes."""
    image = wb_100x100
    fragment = Image.new('RGBA', (10, 10), (0, 0, 0, 0))
    draw = ImageDraw.Draw(fragment)
    draw.rectangle([(0, 0), (5, 5)], (255, 0, 0, 255))
    draw.rectangle([(5, 0), (10, 5)], (255, 0, 0, 127))

    pl.place_fragment(image, fragment, (0, 0, 20, 20), resize_mode)
    points = [(4, 4), (8, 4), (15, 5), (1, 8), (4, 8), (1, 15), (4, 15)]
    for p, ec in zip(points, expect_colors):
        assert image.getpixel(p) == ec


@pytest.fixture()
def recipe(regdir, fragdir):
    """One simple rendering recipe."""
    return pl.Recipe(
        os.path.join(str(regdir), 'white2.png'),
        [
            (os.path.join(str(fragdir), 'red.gif'),
                (0, 0, 10, 20), 'scale'),
            (os.path.join(str(fragdir), 'green.png'),
                (20, 20, 50, 60), 'scale'),
            (os.path.join(str(fragdir), 'subdir', 'orange.jpg'),
                (5, 5, 10, 10), 'crop'),
            (None, (50, 50, 60, 60), 'crop'),  # Don't place anything here.
        ],
    )


def test_render_recipe(tmpdir, recipe):
    target_path = str(tmpdir.join('out.png'))
    recipe.render(target_path)
    output = Image.open(target_path)
    assert output.getpixel((1, 1)) == cc.RED
    assert output.getpixel((21, 21)) == cc.GREEN
    assert output.getpixel((7, 7)) == cc.ORANGE


def test_render_recipes(tmpdir, regdir, recipe):
    src_index = idx.reg_index(str(regdir))
    target_dir = tmpdir.mkdir('td')
    recipes = [recipe, recipe]
    pl.render_recipes(src_index, recipes, str(target_dir))

    # Check that the files were rendered and the index created.
    assert target_dir.join('00000.png').check(file=1)
    assert target_dir.join('00001.png').check(file=1)
    assert target_dir.join('regions.csv').check(file=1)

    # Check index correctness.
    tgt_index = idx.reg_index(str(target_dir))
    assert set(tgt_index) == {'00000.png', '00001.png'}
    assert tgt_index['00000.png'] == src_index['white2.png']
    assert tgt_index['00001.png'] == src_index['white2.png']



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Indexing images in directories."""

import csv
import functools
import glob
import os

from PIL import Image

IMAGE_EXTENSIONS = ['.jpg', '.jpeg', '.png', '.gif']


class Index(dict):
    """Index is a mapping of file names to metadata."""

    def __init__(self, root_path, mapping={}):
        super().__init__(mapping)
        self.root_path = root_path

    @functools.lru_cache(maxsize=32)
    def load_image(self, image_name):
        """Load image from the index."""
        image_path = os.path.join(self.root_path, image_name)
        return Image.open(image_path)

    def add_image(self, image_name, image, metadata=None):
        """Add image into the index."""
        if image_name in self:
            raise KeyError('The index already contains {}'.format(image_name))
        image_path = os.path.join(self.root_path, image_name)
        image.save(image_path)
        self[image_name] = metadata

    def __hash__(self):
        return hash(self.root_path)


class FragmentIndex(Index):
    """Index of a directory with fragments.

    Maps image names/paths to tuples: (width, height).

    """

    def load(self):
        """Load the index from .root_path."""
        for dirpath, dirnames, filenames in os.walk(self.root_path):
            for filename in filenames:
                if any(filename.endswith(ext) for ext in IMAGE_EXTENSIONS):
                    image_path = os.path.join(dirpath, filename)
                    rel_path = os.path.relpath(image_path, self.root_path)
                    image = Image.open(image_path)
                    self[rel_path] = image.size

    def add_image(self, image_name, image):
        """Write image into the index and add it to the index."""
        super().add_image(image_name, image, image.size)


def frag_index(root_path):
    """Return a fragment index of all images under root_path as a dict."""
    ret = FragmentIndex(root_path)
    ret.load()
    return ret


def _clip(x, upper_bound):
    """Return x clipped betweed 0 and upper_bound."""
    if x < 0:
        x = 0
    if x > upper_bound:
        x = upper_bound
    return x


class RegionIndex(Index):
    """Index of a directory with images that have marked regions.

    Maps image names/paths to lists of tuples: (x1, y1, x2, y2, region_type).

    """

    region_types = None  # List of region type names.

    def load(self):
        """Load the index from .root_path."""
        self._load_region_types()
        self._scan_for_images()
        have_csv = self._load_csv_regions()
        have_txt = self._load_yolo_regions()

        if self.region_types is None:
            # Without a provided region types mapping the naming of regions
            # loaded from CSV and TXT files will not be consistent.
            if have_csv and have_txt:
                raise Exception('A .names file is required to allow mixing of '
                                '.csv and .txt regions files.')
            self.region_types = sorted({
                region_type
                for regions in self.values()
                for x1, y1, x2, y2, region_type in regions
            })

        self._clip_regions()

    def add_image(self, image_name, image, regions=[]):
        """Write image into the index root and add regions to the index."""
        super().add_image(image_name, image, regions)

        width, height = image.size
        new_rt = False
        mapped_regions = []

        for x0, y0, x1, y1, rt in regions:
            if rt not in self.region_types:
                self.region_types.append(rt)
                new_rt = True
            t = self.region_types.index(rt)
            x = (x0 + x1) / 2 / width
            y = (y0 + y1) / 2 / height
            w = (x1 - x0) / width
            h = (y1 - y0) / height
            mapped_regions.append((t, x, y, w, h))

        if new_rt:
            self._save_region_types()

        if mapped_regions:
            txt_name = os.path.splitext(image_name)[0] + '.txt'
            txt_path = os.path.join(self.root_path, txt_name)
            with open(txt_path, 'wt', encoding='utf-8') as f:
                for mr in mapped_regions:
                    print(*mr, file=f)

    @functools.lru_cache()
    def _get_image_size(self, image_name):
        """Return image size in pixels as a tuple: (width, height)."""
        image = self.load_image(image_name)
        return image.size

    def _region_types_path(self):
        """Find a .names file inside root_path."""
        names_files = glob.glob(os.path.join(self.root_path, '*.names'))
        if len(names_files) > 1:
            raise Exception('Multiple .names files in {}'
                            .format(self.root_path))
        if len(names_files) == 1:
            return names_files[0]
        return None

    def _load_region_types(self):
        """Load region types from a .names file inside root_path."""
        rt_path = self._region_types_path()
        if rt_path is not None:
            with open(rt_path, 'rt', encoding='utf-8') as f:
                self.region_types = [line.strip() for line in f]

    def _save_region_types(self):
        """Save region types to .names file inside root_path."""
        rt_path = self._region_types_path()
        if rt_path is None:
            rt_path = os.path.join(self.root_path, 'class.names')
        with open(rt_path, 'wt', encoding='utf-8') as f:
            for rt in self.region_types:
                print(rt, file=f)

    def _scan_for_images(self):
        """Find images inside of .root_path and add them to the index."""
        for dirpath, dirnames, filenames in os.walk(self.root_path):
            for filename in filenames:
                file_path = os.path.join(dirpath, filename)
                rel_path = os.path.relpath(file_path, self.root_path)
                if any(filename.endswith(ext) for ext in IMAGE_EXTENSIONS):
                    self[rel_path] = []

    def _clip_regions(self):
        """Clip regions to the image boundaries and remove 0-area ones."""
        for image_name in list(self.keys()):
            width, height = self._get_image_size(image_name)
            clipped_regions = []
            for x1, y1, x2, y2, region_type in self[image_name]:
                x1 = _clip(x1, width)
                x2 = _clip(x2, width)
                y1 = _clip(y1, height)
                y2 = _clip(y2, height)
                if x1 != x2 and y1 != y2:
                    clipped_regions.append((x1, y1, x2, y2, region_type))
            self[image_name] = clipped_regions

    def _load_csv_regions(self):
        """Load regions from any .csv files in root_path."""
        csv_files = glob.glob(os.path.join(self.root_path, '*.csv'))
        for csv_file in csv_files:
            with open(csv_file, 'r', encoding='utf-8') as f:
                reader = csv.DictReader(f)
                for row in reader:
                    image_name = row['image']
                    x1, y1, x2, y2 = [
                        int(round(float(row[key])))
                        for key in ['xmin', 'ymin', 'xmax', 'ymax']
                    ]
                    region_type = row['label']
                    self[image_name].append((x1, y1, x2, y2, region_type))
        return csv_files != []

    def _load_yolo_regions(self):
        """Load regions from image_name.txt files."""
        found_some = False

        for image_name in list(self.keys()):
            txt_name = os.path.splitext(image_name)[0] + '.txt'
            txt_path = os.path.join(self.root_path, txt_name)
            if not os.path.exists(txt_path):
                continue

            if self[image_name] != []:
                raise Exception('Regions specified twice for {}'
                                .format(image_name))
            found_some = True

            width, height = self._get_image_size(image_name)
            with open(txt_path, 'rt', encoding='utf-8') as f: 
                for line in f:
                    t, x, y, w, h = line.strip().split()

                    t = int(t)
                    if self.region_types is not None:
                        t = self.region_types[t]
                    else:
                        t = str(t)

                    x, y, w, h = map(float, [x, y, w, h])
                    x1 = int(round((x - w/2) * width))
                    x2 = int(round((x + w/2) * width))
                    y1 = int(round((y - h/2) * height))
                    y2 = int(round((y + h/2) * height))

                    self[image_name].append((x1, y1, x2, y2, t))

        return found_some


def reg_index(root_path, ensure_dir=False):
    """Return the index of marked regions in the images.

    The directory is scanned for image files and the regions map is loaded from
    CSV files in it.

    Parameters
    ----------
    root_path : str
        Path to the directory from which the index is loaded.
    ensure_dir : bool
        If this is True, the root directory will be created if it doesn't
        exist.

    Returns
    -------
    index : RegionIndex
        Index that maps image filenames to lists of tuples that contain
        bounding boxes and region types.

    """
    if ensure_dir:
        os.makedirs(root_path, exist_ok=True)
    ret = RegionIndex(root_path)
    ret.load()
    return ret


def some_index(root_path):
    """Return region index if possible, otherwise fragment index.

    Checks if there are any .csv or .txt files in the directory. If yes,
    attempt RegionIndex, otherwise FragmentIndex.

    """
    csv_files, txt_files = [
        glob.glob(os.path.join(root_path, ext))
        for ext in ['*.csv', '*.txt']
    ]
    if csv_files or txt_files:
        return reg_index(root_path)
    return frag_index(root_path)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for common utilities."""

import pytest

import admincer.find as fnd
import admincer.util as util


@pytest.mark.parametrize('source,parsed', [
    ('foo:10x10', fnd.RegionQuery('foo', 10, 10)),
    ('bar:100x100:50', fnd.RegionQuery('bar', 100, 100, 50)),
    ('broken', None),
    ('broken:100', None),
    ('broken:cxz', None),
    ('broken:10x10:10:5', None),
])
def test_region_parser(source, parsed):
    if parsed is None:
        with pytest.raises(Exception):
            util.region(source)
    else:
        got = util.region(source)
        assert got.__dict__ == parsed.__dict__


@pytest.mark.parametrize('source,parsed', [
    ('foo/bar.png', fnd.FragmentQuery('foo/bar.png')),
    ('/a/b/c.jpg:42', fnd.FragmentQuery('/a/b/c.jpg', 42)),
    ('broken:1:2', None),
    ('broken:foo', None),
])
def test_fragment_parser(source, parsed):
    if parsed is None:
        with pytest.raises(Exception):
            util.fragment(source)
    else:
        got = util.fragment(source)
        assert got.__dict__ == parsed.__dict__



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for the CLI."""

import fnmatch
import glob
import os
import random

from PIL import Image, ImageDraw
import pytest

import admincer.index as idx

from conftest import make_image, Colors as cc
from test_slice import SLICE_EXPECT, check_results


def test_noargs(script_runner):
    """Show brief usage info when called with no arguments."""
    ret = script_runner.run('admincer')
    assert not ret.success
    assert 'Batch image editor' in ret.stdout
    assert ret.stderr == ''


@pytest.fixture()
def fragdir2(tmpdir):
    """Two more directories with fragments."""
    ret = tmpdir.mkdir('fragdir2')
    make_image(ret.mkdir('1').join('blue.gif'), 10, 10, cc.BLUE)
    make_image(ret.mkdir('2').join('pink.gif'), 10, 10, cc.PINK)
    return ret


@pytest.fixture()
def regdir2(tmpdir):
    """One image with several regions for test_place()."""
    ret = tmpdir.mkdir('regdir2')
    make_image(ret.join('img.png'), 200, 200, cc.WHITE)
    ret.join('index.csv').write('\n'.join([
        'image,xmin,ymin,xmax,ymax,label',
        'img.png,0,0,10,24,a',
        'img.png,10,0,40,40,a',
        'img.png,40,0,90,60,a',
        'img.png,0,100,10,110,b'
    ]))
    return ret


@pytest.fixture()
def check_log(logs_dir):
    """Check logs against expected outputs with filename pattern matching."""

    def checker(output, log_name):
        """Compare output to lines read from log_name with fnmatch."""
        expect_output = logs_dir.join(log_name).read().splitlines()
        got_output = output.splitlines()
        assert len(expect_output) == len(got_output)
        for got, expect in zip(got_output, expect_output):
            assert fnmatch.fnmatch(got, expect)

    return checker


def test_place(script_runner, tmpdir, regdir2, fragdir, fragdir2):
    """Place fragments onto a source image."""
    targetdir = tmpdir.join('output')
    ret = script_runner.run(
        'admincer', 'place',
        '--fragments', 'a=' + str(fragdir),
        # Test providing multiple directories for the same region type.
        '-f', 'b={}:{}'.format(fragdir2.join('1'), fragdir2.join('2')),
        '-n', '10',
        str(regdir2), str(targetdir),
    )
    assert ret.success
    assert ret.stdout == ''
    assert ret.stderr == ''
    outputs = [
        Image.open(img_path)
        for img_path in glob.glob(os.path.join(str(targetdir), '*.png'))
    ]
    # Identify unique combinations of colors in the four regions.
    signatures = {
        tuple(o.getpixel(p) for p in [(1, 22), (11, 1), (41, 1), (1, 101)])
        for o in outputs
    }
    # Since the first three regions only have one candidate fragment, there are
    # only two possibilities for the complete signature.
    assert signatures == {
        (cc.RED, cc.GREEN, cc.ORANGE, cc.BLUE),
        (cc.RED, cc.GREEN, cc.ORANGE, cc.PINK),
    }
    assert targetdir.join('regions.csv').check(file=1)


def test_place_verbose(script_runner, tmpdir, regdir2, fragdir, fragdir2,
                       check_log):
    """Place fragments with verbose output."""
    targetdir = tmpdir.join('output')
    random.seed(0)  # To prevent variation in placement order.
    ret = script_runner.run(
        'admincer', 'place',
        '--verbose',
        '-f', 'a=' + str(fragdir),
        '-f', 'b=' + str(fragdir),
        '-n', '2',
        str(regdir2), str(targetdir),
    )
    assert ret.success
    assert ret.stdout == ''
    check_log(ret.stderr, 'place-v.log')


@pytest.fixture()
def regdir3(tmpdir):
    """Image with three regions of the same size."""
    ret = tmpdir.mkdir('regdir3')
    make_image(ret.join('img.png'), 12, 12, cc.WHITE)
    ret.join('index.csv').write('\n'.join([
        'image,xmin,ymin,xmax,ymax,label',
        'img.png,0,0,10,10,a',
        'img.png,0,0,8,8,b',
        'img.png,0,0,12,12,c',
    ]))
    return ret


def test_place_resize_modes(script_runner, tmpdir, regdir3, fragdir2):
    """Place fragments with different resize modes."""
    targetdir = tmpdir.join('output')
    ret = script_runner.run(
        'admincer', 'place',
        '-f', 'a=' + str(fragdir2),
        '-f', 'b=' + str(fragdir2),
        '-f', 'c=' + str(fragdir2),
        '--resize-mode', 'scale',
        '--resize-mode', 'b=pad',
        '--resize-mode', 'c=crop',
        str(regdir3), str(targetdir),
    )
    assert ret.success
    assert ret.stdout == ''
    assert ret.stderr == ''


def test_invalid_resize_mode(script_runner, tmpdir, regdir3, fragdir2):
    ret = script_runner.run(
        'admincer', 'place',
        '-f', 'a=' + str(fragdir2),
        '-f', 'b=' + str(fragdir2),
        '-f', 'c=' + str(fragdir2),
        '-r', 'foobar',
        str(regdir3), str(tmpdir),
    )
    assert not ret.success
    assert ret.stdout == ''
    assert 'must be "scale", "pad" or "crop"' in ret.stderr


def test_extract(script_runner, tmpdir, regdir):
    """Extract regions into a separate directory."""
    r0 = tmpdir.join('r0')
    r1 = tmpdir.join('r1')
    r2 = tmpdir.join('r2')
    ret = script_runner.run(
        'admincer', 'extract',
        '--target-dir', 'r1=' + str(r1),
        '-t', 'r2=' + str(r2),
        '-t', 'r0=' + str(r0),
        str(regdir),
    )

    assert ret.success
    assert ret.stdout == ''
    assert ret.stderr == ''

    # Check that output files were created. We don't check the content of the
    # files here because there's test_extract.py.
    assert r0.check(dir=1)
    assert len(list(r0.visit())) == 0
    assert len(list(r1.visit())) == 3
    assert len(list(r2.visit())) == 1


def test_extract_debug(script_runner, tmpdir, regdir, check_log):
    """Extract regions with verbose output."""
    r0 = tmpdir.join('r0')
    r1 = tmpdir.join('r1')
    r2 = tmpdir.join('r2')
    ret = script_runner.run(
        'admincer', 'extract',
        '-vv',
        '-t', 'r0=' + str(r0),
        '-t', 'r1=' + str(r1),
        '-t', 'r2=' + str(r2),
        str(regdir),
    )

    assert ret.success
    assert ret.stdout == ''
    check_log(ret.stderr, 'extract-vv.log')


def test_find_region(script_runner, regdir):
    """Find screenshots with marked regions of specific type and size."""
    ret = script_runner.run(
        'admincer', 'find',
        '--region=r1:10x10',
        '-r', 'r3:30x100:100',
        str(regdir),
    )

    assert ret.success
    assert ret.stdout == 'blue3.png\n'


def test_find_fragment(script_runner, fragfind_dir):
    """Find screenshots with specific fragments."""
    needles_dir = fragfind_dir.join('needles')
    ret = script_runner.run(
        'admincer', 'find',
        '--fragment={}'.format(needles_dir.join('2.png')),
        '-f', '{}:5'.format(needles_dir.join('3.png')),
        str(fragfind_dir.join('haystack')),
    )

    assert ret.success
    assert ret.stdout == '1.jpg\n'


def test_find_region_no_index(script_runner, fragfind_dir):
    """Error when --region is given but there's no region index."""
    needles_dir = fragfind_dir.join('needles')
    ret = script_runner.run(
        'admincer', 'find',
        '-r', 'r1:30x100',
        str(needles_dir),
    )

    assert not ret.success
    assert 'region index' in ret.stderr


def test_find_none(script_runner, regdir):
    """Find with no query produces an error."""
    ret = script_runner.run('admincer', 'find', str(regdir))
    assert not ret.success
    assert 'requires at least one' in ret.stderr


def test_slice(script_runner, slice_source, tmpdir):
    """Test slice sub-command."""
    target = tmpdir.join('output')
    ret = script_runner.run(
        'admincer', 'slice', '--no-empty', '--min-part', '80',
        str(slice_source), str(target)
    )
    assert ret.success
    assert ret.stdout == ''
    assert ret.stderr == ''

    results = idx.reg_index(str(target))
    check_results(results, SLICE_EXPECT['no_empty_80'])


def test_slice_verbose(script_runner, slice_source, tmpdir, check_log):
    """Test slice sub-command."""
    target = tmpdir.join('output')
    ret = script_runner.run(
        'admincer', 'slice', '-v', '--step', '10',
        str(slice_source), str(target)
    )

    assert ret.success
    assert ret.stdout == ''
    check_log(ret.stderr, 'slice-v.log')


def test_convert_one(script_runner, convert_src):
    ret = script_runner.run('admincer', 'convert',
                            str(convert_src.join('14_task11.xml')))

    assert ret.success
    results = idx.reg_index(str(convert_src.join('task11')))
    assert results['1_cut_1.png'] == [(976, 461, 1278, 711, 'textad')]
    assert results['1_cut_0.png'] == [(233, 714, 962, 805, 'bannerad')]


def test_convert_multiple_names(script_runner, convert_src_names):
    convert_src_names.join('whatever.names').write('otherad')
    ret = script_runner.run(
        'admincer', 'convert',
        str(convert_src_names.join('14_task11.xml')),
        str(convert_src_names.join('15_task12.xml'))
    )

    assert ret.success
    results11 = idx.reg_index(str(convert_src_names.join('task11')))
    results12 = idx.reg_index(str(convert_src_names.join('task12')))
    assert results11['1_cut_0.png'] == [(233, 714, 962, 805, 'bannerad')]
    assert results12['5_cut_1.png'] == [(975, 257, 1219, 468, 'popupad')]

    names = 'otherad\nbannerad\ntextad\npopupad'
    assert convert_src_names.join('whatever.names').read() == names
    assert convert_src_names.join('task12', 'class.names').read() == names


def test_convert_and_move(script_runner, convert_src):
    ret = script_runner.run('admincer', 'convert',
                            str(convert_src.join('14_task11.xml')),
                            str(convert_src.join('15_task12.xml')),
                            '-t', str(convert_src.join('converted')), '-m')
    assert ret.success
    results = idx.reg_index(str(convert_src.join('converted')))
    assert results['1_cut_1.png'] == [(976, 461, 1278, 711, 'textad')]
    assert results['1_cut_0.png'] == [(233, 714, 962, 805, 'bannerad')]

    for fn in ['1_cut_0.png', '1_cut_1.png', 'no_ads.png']:
        assert convert_src.join('converted', fn).check(exists=1)
    for fn in ['1_cut_0.png', '1_cut_1.png']:
        assert convert_src.join('task11', fn).check(exists=0)


def test_copy_and_move(script_runner, convert_src):
    """Trying to convert with both 'copy' and 'move' should fail."""
    ret = script_runner.run('admincer', 'convert',
                            str(convert_src.join('15_task12.xml')),
                            '-m', '-c')

    assert not ret.success
    assert 'Choose either "copy" or "move", not both' in ret.stderr



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for slicing."""

import pytest

import admincer.index as idx
import admincer.slice as slc

from conftest import image_to_chars, SLICE_STRINGS


SLICE_EXPECT = {}

SLICE_EXPECT['default'] = {
    '0-0.png': {
        'chars': SLICE_STRINGS[0],
        'regions': [(0, 0, 10, 10, 'r1'), (20, 10, 30, 30, 'r2')],
    },
    '1-0.png': {
        'chars': SLICE_STRINGS[1],
        'regions': [],
    },
    '2-0.png': {
        'chars': SLICE_STRINGS[2][:9],
        'regions': [(0, 0, 10, 10, 'r1'), (20, 0, 30, 20, 'r3')],
    },
    '2-10.png': {
        'chars': SLICE_STRINGS[2][3:],
        'regions': [(20, 0, 30, 10, 'r3'), (20, 20, 30, 30, 'r1')],
    },
    '3-0.png': {
        'chars': SLICE_STRINGS[3][:9],
        'regions': [(20, 10, 30, 30, 'r1')],
    },
    '3-10.png': {
        'chars': SLICE_STRINGS[3][3:12],
        'regions': [(20, 0, 30, 20, 'r1')],
    },
    '3-20.png': {
        'chars': SLICE_STRINGS[3][6:15],
        'regions': [(20, 0, 30, 10, 'r1')],
    },
    '3-30.png': {
        'chars': SLICE_STRINGS[3][9:18],
        'regions': [],
    },
    '3-40.png': {
        'chars': SLICE_STRINGS[3][12:21],
        'regions': [(20, 20, 30, 30, 'r2')],
    },
    '3-50.png': {
        'chars': SLICE_STRINGS[3][15:],
        'regions': [(20, 10, 30, 30, 'r2')],
    },
}

SLICE_EXPECT['no_empty'] = {
    name: SLICE_EXPECT['default'][name]
    for name in [
        '0-0.png',
        '2-0.png', '2-10.png',
        '3-0.png', '3-10.png', '3-20.png', '3-40.png', '3-50.png',
    ]
}

SLICE_EXPECT['no_empty_80'] = {
    name: SLICE_EXPECT['default'][name]
    for name in ['0-0.png', '2-0.png', '3-0.png', '3-10.png', '3-50.png']
}
SLICE_EXPECT['no_empty_80']['2-10.png'] = {
    'chars': SLICE_STRINGS[2][3:],
    'regions': [(20, 20, 30, 30, 'r1')],
}

SLICE_EXPECT['step_30'] = {
    name: SLICE_EXPECT['default'][name]
    for name in ['0-0.png', '1-0.png', '2-0.png', '3-0.png', '3-30.png']
}


def check_results(results, expect):
    """Compare the contents of target index with expectations."""
    assert set(results) == set(expect)
    for img_name in results:
        image = results.load_image(img_name)
        assert image.size == (30, 30)
        assert image_to_chars(image, 10, 3, 3) == expect[img_name]['chars']
        assert results[img_name] == expect[img_name]['regions']


@pytest.mark.parametrize('expect_key,kw', [
    ('default', {}),
    ('no_empty', {'no_empty': True}),
    ('no_empty_80', {'no_empty': True, 'min_part': 0.8}),
    ('step_30', {'step': 30}),
])
def test_slice(slice_source, tmpdir, expect_key, kw):
    """Test slice sub-command."""
    target = tmpdir.mkdir('output')
    source = idx.reg_index(str(slice_source))
    output = idx.reg_index(str(target))
    slc.slice_all(source, output, **kw)
    check_results(output, SLICE_EXPECT[expect_key])


def test_double_slice_fail(slice_source, tmpdir):
    target = tmpdir.mkdir('output')
    source = idx.reg_index(str(slice_source))
    output = idx.reg_index(str(target))
    output['0-0.png'] = []
    with pytest.raises(KeyError):
        slc.slice_all(source, output)



# This file is part of AdMincer,
# Copyright (C) 2019-present eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Tests for converting annotations from CVAT to YOLO format."""


import pytest

import admincer.convert as co
import admincer.index as idx


def test_process_cvat_no_output_dir(convert_src):
    co.process_cvat_xml(str(convert_src.join('14_task11.xml')))

    results = idx.reg_index(str(convert_src.join('task11')))
    assert results['1_cut_1.png'] == [(976, 461, 1278, 711, 'textad')]
    assert results['1_cut_0.png'] == [(233, 714, 962, 805, 'bannerad')]
    names = 'bannerad\ntextad'
    assert convert_src.join('task11', 'class.names').read() == names


def test_process_cvat_with_output_dir_and_copy(convert_src):
    co.process_cvat_xml(str(convert_src.join('14_task11.xml')),
                        (str(convert_src.join('converted'))), False, True)

    results = idx.reg_index(str(convert_src.join('converted')))
    assert results['1_cut_0.png'] == [(233, 714, 962, 805, 'bannerad')]
    assert convert_src.join('converted', '1_cut_0.png').check(exists=1)
    assert convert_src.join('task11', '1_cut_0.png').check(exists=1)
    assert convert_src.join('converted', '1_cut_1.png').check(exists=1)
    assert convert_src.join('task11', '1_cut_1.png').check(exists=1)
    assert convert_src.join('converted', 'no_ads.png').check(exists=1)


def test_process_cvat_with_outputdir_names_copy(convert_src_names):
    convert_src_names.join('converted', 'foo.names').write('bar', ensure=True)

    co.process_cvat_xml(str(convert_src_names.join('14_task11.xml')),
                        str(convert_src_names.join('converted')), False, True)
    co.process_cvat_xml(str(convert_src_names.join('15_task12.xml')),
                        str(convert_src_names.join('converted')), False, True)

    results = idx.reg_index(str(convert_src_names.join('converted')))
    assert results['2_cut_0.png'] == [(957, 807, 1262, 1348, 'bannerad')]
    assert results['5_cut_1.png'] == [(975, 257, 1219, 468, 'popupad')]
    assert convert_src_names.join('converted', '2_cut_0.png').check(exists=1)
    assert convert_src_names.join('task12', '2_cut_0.png').check(exists=1)
    assert convert_src_names.join('converted', '5_cut_1.png').check(exists=1)
    assert convert_src_names.join('task12', '5_cut_1.png').check(exists=1)

    names = 'bar\nbannerad\ntextad\npopupad'
    assert convert_src_names.join('converted', 'foo.names').read() == names
    assert convert_src_names.join('task12', 'class.names').read() == names


def test_image_missing(bad_convert_src, caplog):
    co.process_cvat_xml(str(bad_convert_src.join('14_task11.xml')))
    assert 'image cannot be found.' in caplog.records[0].msg

    results = idx.reg_index(str(bad_convert_src.join('task11')))
    assert results['1_cut_0.png'] == [(233, 714, 962, 805, 'bannerad')]


def test_convert_2names(convert_src_names, caplog):
    convert_src_names.join('task11', 'other.names').write('foo')

    with pytest.raises(Exception) as info:
        co.process_cvat_xml(
            str(convert_src_names.join('14_task11.xml')),
            str(convert_src_names.join('converted')), False, True)

    assert str(info.value).startswith('Multiple .names files')



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

import os
from setuptools import setup

README_PATH = os.path.join(os.path.dirname(__file__), 'README.md')

with open(README_PATH) as f:
    long_description = f.read()

setup(
    name='admincer',
    use_scm_version=True,
    description='Tool for managing datasets for visual ad detection',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='eyeo GmbH',
    author_email='info@adblockplus.org',
    url='https://gitlab.com/eyeo/machine-learning/admincer/',
    packages=['admincer'],
    entry_points={
        'console_scripts': ['admincer=admincer.__main__:main'],
    },
    include_package_data=True,
    install_requires=[
        'Pillow',
        'opencv-python',
        'setuptools_scm',
    ],
    setup_requires=[
        'setuptools_scm',
    ],
    license='GPLv3',
    zip_safe=False,
    keywords='ad-detection machine-learning dataset',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Utilities',
    ],
)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""CLI entry point."""

import argparse
import logging
import os
import sys

import admincer.index as idx
import admincer.extract as ex
import admincer.find as fnd
import admincer.place as pl
import admincer.slice as slc
import admincer.convert as co
import admincer.util as util

__all__ = ['main']

parser = argparse.ArgumentParser(description='Batch image editor')
subparsers = parser.add_subparsers(help='sub-command help')


def command(*args, **kw):
    """Return a decorator for command functions.

    This decorator will create a subparser for the command function passing
    all the arguments of `command()` to `.add_parser()`. If no name and help
    are provided for the command, they will be taken from the function name
    and docstring (only the first line of the docstring is used) of the
    decorated function.

    """

    def decorator(func):
        nonlocal args, kw

        if not args:
            args = [func.__name__]
        if 'help' not in kw:
            kw['help'] = func.__doc__.split('\n')[0]

        cmd = subparsers.add_parser(*args, **kw)
        for arg in reversed(getattr(func, '__args__', [])):
            cmd.add_argument(*arg['args'], **arg['kw'])
        cmd.set_defaults(func=func)

        return func

    return decorator


def arg(*args, **kw):
    """Return a decorator that will add an argument to a command function.

    All parameters passed to the decorator will be passed to `.add_argument()`
    call of the subparser corresponding to the decorated function.

    """

    def decorator(func):
        nonlocal args, kw

        if not hasattr(func, '__args__'):
            func.__args__ = []
        func.__args__.append({'args': args, 'kw': kw})

        return func

    return decorator


def verbose_arg():
    """Return a decorator for --verbose option."""
    return arg('--verbose', '-v', action='count', default=0,
               help='Output markdown-formatted log of actions to stderr')


def parse_xy_options(argitems, multivalue=False):
    """Collect options of the form x=y or x=y1:y2 into a list of tuples."""
    ret = []

    for s in argitems:
        if '=' in s:
            x, y = s.split('=')
        else:
            x, y = None, s

        if multivalue:
            y = y.split(':')

        ret.append((x, y))

    return ret


def take(n, iterable):
    """Return an iterable with n first items of the original iterable."""
    for i, item in enumerate(iterable):
        yield item
        if i >= n - 1:
            return


@command(aliases=['pl'])
@arg('source', help='Directory for source images and region maps')
@arg('target', help='Directory for output images')
@arg('--fragments', '-f', action='append', default=[],
    metavar='REGION-TYPE=DIR[:DIR...]',
    help='Add fragment directory/ies for specific region types')
@arg('--resize-mode', '-r', action='append', default=[],
    metavar='REGION-TYPE=MODE',
    help='Resize mode for region types: "scale" (default), "pad" or "crop"')
@arg('--count', '-n', default=1, type=int, help='Number of images to generate')
@verbose_arg()
def place(args):
    """Place fragments into designated regions of source images."""
    frag_dirs = parse_xy_options(args.fragments, multivalue=True)

    logging.info('# Place')
    logging.info('- source dir: %s', args.source)
    logging.info('- fragments:')
    for rt, dirs in frag_dirs:
        logging.info('  - %s: %s', rt, ', '.join(dirs))
    logging.info('- target dir: %s', args.target)

    reg_index = idx.reg_index(args.source)
    frag_indices = [(rt, map(idx.frag_index, dirs)) for rt, dirs in frag_dirs]
    resize_modes = dict(parse_xy_options(args.resize_mode))
    recipe_gen = pl.gen_recipes(reg_index, frag_indices, resize_modes)
    recipes = take(args.count, recipe_gen)
    pl.render_recipes(reg_index, recipes, args.target)


@command(aliases=['sl'])
@arg('source', help='Directory for source images and region maps')
@arg('target', help='Directory for output images and region maps')
@arg('--step', '-s', default=10, type=int,
     help='Number of pixels to shift the viewport between the slices')
@arg('--min-part', '-p', default=50, type=int,
     help='Minimal percentage of the region in the slice')
@arg('--no-empty', '-n', action='store_true',
     help='Skip slices without regions')
@verbose_arg()
def slice(args):
    """Cut viewport screenshots from full page screenshots."""

    logging.info('# Slice')
    logging.info('- source dir: %s', args.source)
    logging.info('- target dir: %s', args.target)
    logging.info('- slicing step: %d pixels', args.step)
    logging.info('- include regions that overlap the slice by at least %d%%',
                 args.min_part)
    if args.no_empty:
        logging.info('- skip slices with no regions')

    src_index = idx.reg_index(args.source)
    tgt_index = idx.reg_index(args.target, ensure_dir=True)

    slc.slice_all(
        src_index, tgt_index,
        step=args.step,
        min_part=args.min_part / 100,
        no_empty=args.no_empty,
    )


@command(aliases=['co'])
@arg('source', nargs='+',
     help='CVAT .xml file(s) to convert to YOLO')
@arg('--target-dir', '-t',
     help='Directory to place converted images and YOLO .txt annotations')
@arg('--move', '-m', action='store_true',
     help='Option to move the images to the --target-dir')
@arg('--copy', '-c', action='store_true',
     help='Option to copy the images to the --target-dir')
def convert(args):
    """Convert annotations from CVAT to YOLO format."""
    target_dir = os.path.abspath(args.target_dir) if args.target_dir else None

    for xml_file in args.source:
        co.process_cvat_xml(os.path.abspath(xml_file), target_dir,
                            move=args.move, copy=args.copy)


@command(aliases=['ex'])
@arg('source', help='Directory for source images')
@arg('--target-dir', '-t', action='append', default=[],
     metavar='REGION-TYPE=DIR',
     help='Target directory for extracting regions of specific type')
@verbose_arg()
def extract(args):
    """Extract the contents of fragments to directories by type."""
    target_dirs = sorted(dict(parse_xy_options(args.target_dir)).items())

    logging.info('# Extract')
    logging.info('- source dir: %s', args.source)
    logging.info('- target dirs:')

    for rt, d in target_dirs:
        logging.info('  - %s: %s', rt, d)

    reg_index = idx.reg_index(args.source)
    for region_type, target_dir in target_dirs:
        logging.info('## Type: %s to: %s', region_type, target_dir)
        os.makedirs(target_dir, exist_ok=True)
        regions = ex.extract_regions(reg_index, region_type)
        for i, (box, region) in enumerate(regions):
            out_name = '{:05d}.png'.format(i)
            out_path = os.path.join(target_dir, out_name)
            region.save(out_path)
            logging.info('- region %s to: %s', util.format_box(box), out_name)


@command(aliases=['f'])
@arg('dir', help='Directory in which to search')
@arg('--fragment', '-f', action='append', default=[],
     metavar='PATH[:T]', type=util.fragment,
     help='Look for fragments similar to specific image')
@arg('--region', '-r', action='append', default=[],
     metavar='TYPE:WxH[:T]', type=util.region,
     help='Look for regions of specific types and sizes')
def find(args):
    """Find images with regions of specific types and sizes."""
    dir_index = idx.some_index(args.dir)
    queries = args.region + args.fragment
    if not queries:
        sys.exit('find requires at least one of: -r/--region, -f/--fragment')
    for found in fnd.find(dir_index, queries):
        print(found)


def _configure_logging(args):
    """Configure logging."""
    verbosity = getattr(args, 'verbose', 0)

    if verbosity == 1:
        log_level = logging.INFO
    elif verbosity >= 2:
        log_level = logging.DEBUG
    else:
        log_level = logging.WARNING

    logging.basicConfig(stream=sys.stderr, level=log_level,
                        format='%(message)s')


def main():
    """Run the CLI."""
    args = parser.parse_args()
    if callable(getattr(args, 'func', None)):
        _configure_logging(args)
        args.func(args)
    else:
        parser.print_help()
        sys.exit(1)



# This file is part of AdMincer,
# Copyright (C) 2019 eyeo GmbH
#
# AdMincer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# AdMincer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AdMincer. If not, see <http://www.gnu.org/licenses/>.

"""Slice full page screenshots into viewport screenshots."""

import logging
import os

import admincer.util as util


def _slice_boxes(width, height, step):
    """Yield coordinates of possible slices."""
    if width >= height:
        yield (0, 0, width, height)
        return
    for y_offset in range(0, height - width + 1, step):
        yield (0, y_offset, width, y_offset + width)


def _intersect(box1, box2):
    """Compute intersection of two boxes."""
    x0 = max(box1[0], box2[0])
    y0 = max(box1[1], box2[1])
    x1 = min(box1[2], box2[2])
    y1 = min(box1[3], box2[3])

    if x0 >= x1 or y0 >= y1:
        return None

    return (x0, y0, x1, y1) + box1[4:]


def _box_area(box):
    """Compute the area of the box."""
    x0, y0, x1, y1 = box[:4]
    return (x1 - x0) * (y1 - y0)


def _relative_to(box, region):
    """Translate region coordinates to be relative to box."""
    x_org, y_org = box[:2]
    x0, y0, x1, y1 = region[:4]
    return (x0 - x_org, y0 - y_org, x1 - x_org, y1 - y_org) + region[4:]


def _crop_regions(regions, box, min_part):
    """Crop regions to box.
    
    Discards the ones of which less than min_part remains. Translates the
    remaining ones so that their coordinates are relative to the box.
    """
    result = []

    for region in regions:
        intersected = _intersect(region, box)
        if intersected is None:
            continue
        if _box_area(intersected) / _box_area(region) < min_part:
            continue
        result.append(_relative_to(box, intersected))

    return result


def slice_all(src_index, tgt_index, step=10, min_part=0.5, no_empty=False):
    """Slice images with regions into square sub-images.

    Note: at the moment only vertical slicing of tall images is supported.

    Parameters
    ----------
    src_index : RegIndex
        Source of images and region maps are taken.
    tgt_index : RegIndex
        Target index (and directory) for the output images and region maps.
    step : int
        The distance to shift the viewport between slices (in pixels).
    min_part : float
        If a region overlaps the viewport by this percentage, it will be
        included in the slice (appropriately cropped and translated).
    no_empty : bool
        If this is True, slices without regions will be skipped.

    Raises
    ------
    KeyError
        If tgt_index already contains any of the slices.

    """
    for image_name, regions in sorted(src_index.items()):
        image = src_index.load_image(image_name)
        size = image.size
        logging.info('## %s (%dx%d)', image_name, *size)
        width, height = image.size

        for box in _slice_boxes(width, height, step):
            slice_regions = _crop_regions(regions, box, min_part)
            if no_empty and slice_regions == []:
                continue
            base_name = os.path.splitext(image_name)[0]
            slice_name = '{}-{}.png'.format(base_name, box[1])
            slice_image = image.crop(box)

            logging.info('### slice %s', util.format_box(box))
            if slice_regions:
                for region in slice_regions:
                    logging.info('- region %s type %s',
                                 util.format_box(region[:4]), region[4])
            else:
                logging.info('- no regions')
            logging.info('- saved as %s', slice_name)

            tgt_index.add_image(slice_name, slice_image, slice_regions)
